/*
 * Written by Timothy Reluga.
 * Most of work finished in the summer of 2005 for the paper
 * published in BMB.
 *
 * This is a simulation program that simulates an ornstein-uhlenbeck
 * process with reproduction and death, and is used to calculate
 * invasion speeds.
 */
#import <unistd.h>
#import <iostream>
#import "mymath.h"
#import "myassert.h"
#import "option.h"
#import <assert.h>
#import <gsl/gsl_errno.h>
#import <gsl/gsl_randist.h>

using namespace std;

gsl_rng* random_number_generator;

#import "diffusion.h"
#import "uhlenbeck.h"
#import "contact.h"
#import "birthdeath.h"
#import "hyperbolic.h"
#import "twod_uhlenbeck.h"
#import "snapshots.h"

int main(int argc, const char** argv) {

	const char* opt_message[] = {
		"Program: uhlenbeck, June 2, 2005",
		"Started:  September 28, 2004",
		"Purpose: An alternative to a diffusion model",
		"\tUses a C array of double as the main data structure",
		"\tIncludes removal and differentiatable behavior",
		"Warning: hypberbolic code is unconfirmed(March 6, 2005)",
		"Warning: snapshot code is unconfirmed(June 2, 2005)",
		"Usage: uhlenbeck <options>",
		"Options:",
		"\t-v\tdisplays this message.",
		"\t\tModes: {U}hlenbeck",
		"\t\t       {D}iffusion",
		"\t\t       {C}ontact",
		"\t\t       {B}irthdeath",
		"\t\t       {H}yberbolic (unconfirmed)",
		"\t\t       {2}-dimensions (unconfirmed)",
		"\t\t       {J}an's recommended discrete time(snapshots) model",
		NULL
	}; Options opt(opt_message); //opt.exit_on_verbose=false;

	const char diffusion_model  = 'D';
	const char contact_model    = 'C';
	const char uhlenbeck_model  = 'U';
	const char birthdeath_model = 'B';
	const char hyperbolic_model = 'H';
	const char twod_model       = '2';
	const char snapshots_model   = 'J';
	char model_mode = uhlenbeck_model;
	opt.AddOption(new OptionChar('M',&model_mode, (const char*)
		"\t-Mc\tmode of simulation(char, {U,D,C,B,H,2,J})"));
	double r_transmission = 1.0;
	opt.AddOption( new OptionDouble('r', &r_transmission, (const char*)
		"\t-r#\tper-capita reproduction per unit of time(double)"));
	double r_alpha = 1.0;
	opt.AddOption( new OptionDouble('a',&r_alpha, (const char*)
		"\t-a#\tbinding strength of sick(double)"));
	double r_diffusion = 1.0;
	opt.AddOption( new OptionDouble('d',&r_diffusion, (const char*)
		"\t-d#\tdiffusion rate(double)"));
	double r_removal = 0.0;
	opt.AddOption( new OptionDouble('m',&r_removal, (const char*)
		"\t-m#\tremoval rate(double)"));
	int num_particles_max = 1000;
	opt.AddOption( new OptionInt('n',&num_particles_max, (const char*)
		"\t-n#\tmaximum number of particles(int)"));
	double time_step = 1.0;
	opt.AddOption( new OptionDouble('S',&time_step, (const char*)
		"\t-S#\ttime step between observations for Jan's mode(double)"));
	int trials = 1;
	opt.AddOption( new OptionInt('T',&trials, (const char*)
		"\t-T#\tnumber of times to repeat the simulation(int)"));
	int random_number_seed = (int) time(NULL) + ( (int) getpid() << 14  ) ;
	opt.AddOption( new OptionInt('R',&random_number_seed, (const char*)
		"\t-R#\tset the pseudorandom number seed value(int)"));
	double r_width = 1.0;
	opt.AddOption( new OptionDouble('w',&r_width, (const char*)
		"\t-w#\tstandard deviation of healthy from binding site"));
	bool ill_sameas_healthy = false;
	opt.AddOption( new OptionBool('I',&ill_sameas_healthy, (const char*)
		"\t-I\tinfected and healthy individuals behave the same, ignoring -w"));
	bool measureAnchors = false;
	opt.AddOption( new OptionBool('A',&measureAnchors, (const char*)
		"\t-A\tmeasure front by anchor positions instead of exact positions"));
	bool output_parameterization = false;
	opt.AddOption(new OptionBool('P',&output_parameterization, (const char*)
		"\t-P\tprint simulation parameterization first(bool)"));
	opt.CheckOptions(argc, argv);

	Assert(time_step > 0.0, "Time steps must be positive");
	Assert(r_alpha>=0,"Negative values for -a not allowed");
	Assert(r_transmission>=0,"Negative values for -r not allowed");
	Assert(r_removal>=0,"Negative values for -m not allowed");
	Assert(r_diffusion>=0,"Negative values for -d not allowed");
	Assert(r_width>=0,"Negative values for -w not allowed");

	// Should have a flag making this modification optional
	if ( true ) time_step = time_step/r_transmission;

	double r_sigma  = sqrt(r_diffusion/r_alpha);
	if (ill_sameas_healthy) r_width = r_sigma;

	if (output_parameterization) {
		opt.Print();
		printf("# Time, FFF position, population size, trial\n");
	}

	// ----------   Random number Generator setup ---------- //
	gsl_rng_env_setup();
	//gsl_rng * rng = gsl_rng_alloc( gsl_rng_taus );
	//gsl_rng * rng = gsl_rng_alloc( gsl_rng_random64_glibc2 );
	random_number_generator = gsl_rng_alloc( gsl_rng_default );
	gsl_rng_set(random_number_generator, random_number_seed );

	if ( model_mode == diffusion_model ) {
		Assert( not measureAnchors,
			"Diffusion can not be measured by anchors");
		return
		diffusion(
			r_transmission,
			r_removal,
			r_diffusion,
			num_particles_max,
			trials
		);
	} else if ( model_mode == contact_model ) {
		return
		contact(
			r_transmission,
			r_removal,
			r_sigma,
			r_width,
			num_particles_max,
			trials
		);
	} else if ( model_mode == birthdeath_model ) {
		return
		birthdeath(
			r_transmission,
			r_removal,
			num_particles_max,
			trials
		);
	} else if (model_mode == hyperbolic_model){
		return hyperbolic(
			r_transmission,
			r_removal,
			r_width,
			r_alpha,
			num_particles_max,
			trials
		);
	} else if (model_mode == twod_model){
		printf("# 2 dimensional model\n");
		return twod_uhlenbeck(
			r_transmission,
			r_removal,
			r_sigma,
			r_width,
			r_alpha,
			num_particles_max,
			trials,
			measureAnchors
		);
	} else if (model_mode == uhlenbeck_model){
		// Calculate the std of infected individuals
		Assert(r_alpha>0.,"alpha = 0 introduces singularies in uhlenbeck");
		return
		uhlenbeck(
			r_transmission,
			r_removal,
			r_sigma,
			r_width,
			r_alpha,
			num_particles_max,
			trials,
			measureAnchors
		);
	} else if (model_mode == snapshots_model){
		// Calculate the std of infected individuals
		Assert(r_alpha>0.,"alpha = 0 introduces singularies in uhlenbeck");
		return
		snapshots(
			r_transmission,
			r_removal,
			r_sigma,
			r_width,
			r_alpha,
			num_particles_max,
			trials,
			measureAnchors,
			time_step
		);
	} else {
		Assert( false, "-M* was not a valid mode of simulation.");
	}
	return 0;
}
